/*LIBRERIA CONTENENTE LE PROCEDURE PER LA GESTIONE DEL COSTRUTTO MONITOR*/
#include <stdio.h>
#include <sys/sem.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/sem.h>

/* definisco il monitor come un insieme di:
   id_mutex: identificativo del semaforo mutex
   num_cond: numero di variabili condition del monitor
   cond: identificativo del gruppo di semafori relativi alle variabili condition
*/
typedef struct { int id_mutex, num_cond, cond; } monitor;

//una variabile condition  identificata dal numero d'ordine tra quelle di uno stesso monitor
typedef int condition;

/*Prototipi delle funzioni*/
void enter_monitor(monitor *m);    /*ingresso nel monitor*/
void leave_monitor(monitor *m);    /*abbandono del monitor*/
monitor* implem_monitor(int numero_condition);   /*creazione del monitor*/
void remove_monitor(monitor *m);   /*rimozione del monitor*/
void waitcond(monitor *m,condition c); /*wait su una variabile condition*/
void signalcond(monitor *m,condition c);/*signal su una variabile condition*/
int queuecond(monitor m, condition c); /*coda su una variabile condition*/
void waitsem(int id_sem, int num_sem);
void signalsem(int id_sem, int num_sem);

monitor* implem_monitor(int n)
{
	// chiavi private per i gruppi di semafori
	key_t key_mutex=IPC_PRIVATE;
	key_t key_semcond=IPC_PRIVATE;
	monitor *m;
	struct sembuf s; // struttura di appoggio per le perazioni di semop
	int i;

	m=(monitor*) malloc(sizeof(monitor));// riserva la memoria per un monitor
	m->num_cond=n; // settta il numero di variabili condition
	m->id_mutex=semget(key_mutex,1,IPC_CREAT | 0644); //assegna un semaforo per mutex
	m->cond=semget(key_semcond,n,IPC_CREAT | 0644); // assegna n semafori per le condition

	semctl(m->id_mutex,0,SETVAL,1);  /*inizializza MUTEX a verde*/

	for(i=1;i<=n;i++)
      	/*inizializza i semafori associati alle variabili condition a rosso*/
     		semctl(m->cond,i,SETVAL,0);

	return(m);

}

void remove_monitor(monitor *m)
{
  // dealloca i semafori
  semctl(m->id_mutex, 1, IPC_RMID, NULL);
  semctl(m->cond, m->num_cond, IPC_RMID, NULL);
}

void enter_monitor(monitor *m)
{
	waitsem(m->id_mutex,0);
}

void leave_monitor(monitor *m)
{
	signalsem(m->id_mutex,0);
}

void waitcond(monitor *m, condition cond)
{
	signalsem(m->id_mutex,0);
	waitsem(m->cond,cond);
}

void signalcond(monitor *m,condition cond)
{
	if (queuecond(*m,cond)>0)
		signalsem(m->cond,cond);
	else
		signalsem(m->id_mutex,0);
}

int queuecond(monitor m,condition cond)
{
   // restituisce true se ci sono processi in attesa sulla variabile condition cond del monitor m
   return (semctl(m.cond, cond, GETNCNT, NULL)>0);
}

void waitsem(int id_sem, int num_sem)
{
struct sembuf s;

  s.sem_num=num_sem;
  s.sem_op=-1;
  s.sem_flg=SEM_UNDO;
  semop(id_sem, &s, 1);
}

void signalsem(int id_sem, int num_sem)
{
struct sembuf s;

  s.sem_num=num_sem;
  s.sem_op=1;
  s.sem_flg=SEM_UNDO;
  semop(num_sem, &s, 1);
}


